/*
 *  mvPCIe Driver - PCIe implementation
 *  Copyright (c) 2019-2020 by MATRIX VISION GmbH <info@matrix-vision.de>
 *  Author: grayh
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/errno.h>
#include <linux/uaccess.h>
#include <linux/version.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/vmalloc.h>
#include <linux/sysfs.h>

#include "mvdma.h"

static u32 number_of_devices = 0;
static dev_t mvpci_devnum;
static struct kobject *mvpci_kobj;

u32 debug = 0;
u32 isr_debug = 0;

module_param(debug, int, 0644);
MODULE_PARM_DESC(debug, "configure general debug level");
module_param(isr_debug, int, 0644);
MODULE_PARM_DESC(isr_debug, "configure isr debug level");

MODULE_DESCRIPTION(MVPCI_DRIVER_NAME " driver");
MODULE_AUTHOR("MATRIX VISION GmbH <info@matrix-vision.de>");
MODULE_LICENSE("GPL");

//#define USE_NUMA
#define USE_COHERENT_DMA
#define TRY_MSI
//#define TRY_MSIX /* not yet implemented in FPGA */
//#define USE_PCI_ERROR_HANDLERS /* not yet implemented */
#define COWARDLY_REFUSAL

#define _kB (1024)
#define _MB (_kB * _kB)

static struct mvpci_driver mvpci_devices[MVPCI_MAX_DEVICES];

/* ===================================== Interrupts ==========================
 */

#define INT_MSEQ   0
#define INT_IO     1
#define INT_STAT   2
#define INT_SC     3

typedef irqreturn_t (*irq_uio_handler_t)(int, struct uio_info *);

#define DEFINE_UIO_MSI_ISR(n)                               \
static irqreturn_t mvpci_isr_uio ## n (int irq, struct uio_info *info) \
{                                                           \
    struct mvpci_driver *mvpci_drv = (struct mvpci_driver *)info->priv; \
    struct mvpci_fh *fh = &mvpci_drv->fh[0];                \
                                                            \
    /* device not open, ignore interrupt but mark it as processed,  \
     * as it definitely comes from our hardware but cannot be correctly used (yet)  \
     */                                                     \
    if ((number_of_devices == 0) || !fh || (fh->opened == 0)) { \
        return IRQ_RETVAL(1);                               \
    }                                                       \
    iprintk(0x200, "INTERRUPT UIO ISR irq=%d, name=%s\n", irq, info->name);      \
    return IRQ_RETVAL(1);                                   \
}

DEFINE_UIO_MSI_ISR(INT_IO)
DEFINE_UIO_MSI_ISR(INT_STAT)
DEFINE_UIO_MSI_ISR(INT_SC)

#define MSI_UIO_ISR(n) mvpci_isr_uio ## n

/* interrupt handler used for MSeq with MSI and MSI-X */
static irqreturn_t mvpci_isr_mseq (int irq, void *context)
{
    struct mvpci_driver *mvpci_drv = (struct mvpci_driver *)context;
    struct mvpci_fh *fh = &mvpci_drv->fh[0];
    int handled = 0;

    /* device not open, ignore interrupt but mark it as processed,
     * as it definitely comes from our hardware but cannot be correctly used (yet)
     */
    if ((number_of_devices == 0) || !fh || (fh->opened == 0)) {
        return IRQ_RETVAL(1);
    }

    iprintk(0x200, "MSeq INTERRUPT ISR irq=%d\n", irq);
    if (mvdma_irq(irq, context) == IRQ_HANDLED) {
        handled = 1;
    }
    return IRQ_RETVAL(handled);
}

/* interrupt handler used when MSI vectors not possible i.e. one global handler
 * for everything */
static irqreturn_t mvpci_isr_single(int irq, void *context)
{
    int handled = 0;
    struct mvpci_driver *mvpci_drv = (struct mvpci_driver *)context;
    struct mvpci_fh *fh = &mvpci_drv->fh[0];
    u32 irq_source;

    /* device not open, ignore interrupt and mark it as NOT processed,
     * because this might be a genuine shared interrupt from different hardware!
     */
    if ((number_of_devices == 0) || !fh || (fh->opened == 0)) {
        return IRQ_RETVAL(0);
    }

    /* read the source of the interrupt */
    irq_source = mvread(MVPCI_REMEMBERED_IRQS) & 0xF;

    /* if any bit is set it must be one of our interrupts */
    handled = (irq_source > 0);

    if (handled) {
        iprintk(0x200, "irq_source=0x%x\n", irq_source);
        if ( irq_source & IRQ_SOURCE_MSEQ ) {
            /* microsequencer interrupt - call the MSeq handler directly */
            mvpci_isr_mseq(irq, context);
        }

        if ( irq_source & IRQ_SOURCE_1 ) {
            /* I/O interrupt - trigger UIO */
            if (mvpci_drv->uio_info[INT_IO].name) {
                uio_event_notify(&mvpci_drv->uio_info[INT_IO]);
            }
        }

        if ( irq_source & IRQ_SOURCE_2 ) {
            /* statistic interrupt - trigger UIO */
            if (mvpci_drv->uio_info[INT_STAT].name) {
                uio_event_notify(&mvpci_drv->uio_info[INT_STAT]);
            }
        }

        if ( irq_source & IRQ_SOURCE_3 ) {
            /* statistic interrupt - trigger UIO */
            if (mvpci_drv->uio_info[INT_SC].name) {
                uio_event_notify(&mvpci_drv->uio_info[INT_SC]);
            }
        }
    }

    return IRQ_RETVAL(handled);
}

/* ===================================== Helpers =============================
 */
static int set_allocate_name(const char **pname, const char *fmt, ...)
{
    const char *s;
    va_list vargs;

    va_start(vargs, fmt);
    s = kvasprintf_const(GFP_KERNEL, fmt, vargs);
    va_end(vargs);
    if (!s)
        return -ENOMEM;
    *pname = s;
    return 0;
}


/* ============================ Hardware Initialization ==================== */

static int mvpci_reset_hw(struct mvpci_driver *mvpci_drv)
{
    dprintk(MVDMA_DEBUG_STD, "+ %s\n", __FUNCTION__);

    // do something with the HW here

    dprintk(MVDMA_DEBUG_STD, "- %s\n", __FUNCTION__);
    return 0;
}

static int mvpci_init_hw(struct mvpci_driver *mvpci_drv)
{
    int ret = 0;

    dprintk(MVDMA_DEBUG_STD, "+ %s\n", __FUNCTION__);

    mvpci_reset_hw(mvpci_drv);

    dprintk(MVDMA_DEBUG_STD, "- %s\n", __FUNCTION__);

    return ret;
}

/* ===================================== PCI INIT ========================== */

static void mvpci_free_pci(struct mvpci_driver *mvpci_drv)
{
    int i;

    if (!mvpci_drv) {
        return;
    }

    /* remove DMA engine */
    mvdma_remove(mvpci_drv);

    /* unregister MSeq interrupt */
    if (mvpci_drv->msi_x_entry[INT_MSEQ].vector) {
        dprintk(0, "device mvpci-%d: freeing MSI vector %d\n", mvpci_drv->nr, mvpci_drv->msi_x_entry[INT_MSEQ].vector);
        free_irq(mvpci_drv->msi_x_entry[INT_MSEQ].vector, mvpci_drv);
        mvpci_drv->msi_x_entry[INT_MSEQ].vector = 0;
    }

    /* unregister UIO devices, removing interrupts */
    for (i = INT_IO; i <= INT_SC; i++) {
        if (mvpci_drv->uio_info[i].name) {
            if (mvpci_drv->msi_x_entry[i].vector) {
                dprintk(0, "device mvpci-%d: freeing UIO vector %d\n", mvpci_drv->nr, mvpci_drv->msi_x_entry[i].vector);
                uio_unregister_device(&mvpci_drv->uio_info[i]);
                // ... and free the name which we allocated dynamically
                kfree_const(mvpci_drv->uio_info[i].name);
                mvpci_drv->uio_info[i].name = "";
                mvpci_drv->msi_x_entry[i].vector = 0;
            }
        }
    }

    /* unregister all IRQs from device */
    mvpci_drv->flags &= ~(MVPCI_FLAG_IRQ | MVPCI_FLAG_MSI);
    pci_free_irq_vectors(mvpci_drv->pci);

    if (mvpci_drv->iow_sequencer) {
        dprintk(1, "device mvpci-%d: free io_writer sequencer memory\n", mvpci_drv->nr);
        dma_free_coherent(&mvpci_drv->pci->dev, MVPCI_IOW_SEQUENCER_SIZE, mvpci_drv->iow_sequencer, mvpci_drv->dma_iow_sequencer);
        mvpci_drv->iow_sequencer = NULL;
        mvpci_drv->dma_iow_sequencer = 0;
    }

    if (mvpci_drv->mmio) {
        dprintk(1, "device mvpci-%d: pci_iounmap MMIO\n", mvpci_drv->nr);
        pci_iounmap(mvpci_drv->pci, mvpci_drv->mmio);
        mvpci_drv->mmio = NULL;
    }

    if (mvpci_drv->flags & MVPCI_FLAG_RES0) {
        dprintk(1, "device mvpci-%d: pci_release_regions\n", mvpci_drv->nr);
        pci_release_regions(mvpci_drv->pci);
        mvpci_drv->flags &= ~MVPCI_FLAG_RES0;
    }

    pci_set_drvdata(mvpci_drv->pci, NULL);

    dprintk(0, "device mvpci-%d: disable PCI device\n", mvpci_drv->nr);
    pci_disable_device(mvpci_drv->pci);
    mvpci_drv->initialized = 0;
}

static int mvpci_open(struct inode *inode, struct file *file)
{
    unsigned int minor = MINOR(inode->i_rdev);

    dprintk(2, "+ %s, minor = %d\n", __FUNCTION__, minor);

    if (minor < MVPCI_MAX_DEVICES) {
        unsigned n = minor; /* device number */
        struct mvpci_driver *mvpci_drv = &mvpci_devices[n];
        struct mvpci_fh *fh = &mvpci_drv->fh[0];
        struct mvdma_device *pdma = mvpci_drv->pdma;

        file->private_data = fh;
        fh->opened++;
        dprintk(2, "  %s:, device number = %d, opened = %d\n",
            __FUNCTION__, n, fh->opened);

        if (fh->opened == 1) {
            /* DMA should not be running and offset set to 0 if this is the first time opened */
            dprintk(0, "%s: reset queue and descriptor counters to 0\n", __FUNCTION__);
            mvdma_reset(pdma); /* what about the lock GRA ? */
            /* enable Interrupts from the hardware */
            mvwrite(0x04, MVPCI_IO_CONTROL);
        }
    }

    dprintk(2, "- %s\n", __FUNCTION__);
    return 0;
}

static int mvpci_release(struct inode *inode, struct file *file)
{
    dprintk(2, "+ %s\n", __FUNCTION__);

    if (file->private_data) {
        struct mvpci_fh *fh = (struct mvpci_fh *)file->private_data;

        if (fh) {
            struct mvpci_driver *mvpci_drv = fh->mvpci_drv;

            if (fh->opened)
                fh->opened--;

            dprintk(2, "  %s:, device number = %d, opened = %d\n",
                __FUNCTION__, mvpci_drv->nr, fh->opened);

            if (mvpci_drv) {
                struct mvdma_device *pdma = mvpci_drv->pdma;
                if (pdma) {
                    if (fh->opened == 0) {
                        /* DMA should not be running and offset set to 0 if this is the last time closed */
                        dprintk(0, "%s: reset queue and descriptor counters to 0\n", __FUNCTION__);
                        mvdma_reset(pdma); /* what about the lock GRA ? */
                        /* disable Interrupts from the hardware */
                        mvwrite(0x00, MVPCI_IO_CONTROL);
                    }
                }
            }
        }
    }
    file->private_data = NULL;

    dprintk(2, "- %s\n", __FUNCTION__);
    return 0;
}

static ssize_t mvpci_write_seq(struct file *file, const char __user *buf,
               size_t count, loff_t *ppos)
{
    int result = 0;
    u64 addr = (u64)(*ppos) - MVPCI_IOW_SEQUENCER_OFFSET;
    struct mvpci_fh *fh = (struct mvpci_fh *)file->private_data;
    struct mvpci_driver *mvpci_drv = fh->mvpci_drv;

    if ((addr + count) <= MVPCI_IOW_SEQUENCER_SIZE) {
        u8 *p = (u8 *)mvpci_drv->iow_sequencer + addr;
        result = copy_from_user(p, buf, count);
        if (result) {
            pr_err("ERROR: write to addr=0x%llx, count=0x%zx, result=0x%x.\n", addr, count, result);
            return -EINVAL;
        }
    } else { 
        pr_err("ERROR: write length (0x%zx) invalid. Must be <= 0x%llx.\n", count, MVPCI_IOW_SEQUENCER_SIZE);
        return -EINVAL;
    }

#if 0
    dprintk(2, "%s write to addr=0x%llx, count result=0x%x\n", __FUNCTION__, addr, result);
#endif
    return result ? result : count;
}

static ssize_t mvpci_write(struct file *file, const char __user *buf,
               size_t count, loff_t *ppos)
{
    int result = 0;
    u64 addr = (u64)(*ppos);
    u64 val = 0;
    struct mvpci_fh *fh = (struct mvpci_fh *)file->private_data;
    struct mvpci_driver *mvpci_drv = fh->mvpci_drv;

    if (addr >= MVPCI_IOW_SEQUENCER_OFFSET) {
        /* writing to io_writer sequencer memory */
        return mvpci_write_seq(file, buf, count, ppos);
    }

    if (count < sizeof(u32) || count > sizeof(u64)) {
        pr_err("ERROR: write length invalid.\n");
        return -EINVAL;
    }

    /* writing to the BAR */
    if (addr > pci_resource_len(mvpci_drv->pci, 0)) {
        pr_err("ERROR: Offset 0x%llx is outside range of BAR 0 (0x%llx).\n",  addr, pci_resource_len(mvpci_drv->pci, 0));
        return -ENOMEM;
    }

    result = copy_from_user(&val, buf, count);
    if (!result) {
        if (count == sizeof(u32))
            mvwrite(val, addr);
        else if (count == sizeof(u64))
            mvwrite64(val, addr);
#if 0
        dprintk(2, "%s: wrote to addr=0x%llx, val=0x%llx, count=0x%lx\n", __FUNCTION__, addr, val, count);
#endif
    }
    return result ? result : count;
}

static ssize_t mvpci_read_seq(struct file *file, char __user *buf, size_t count,
              loff_t *ppos)
{
    int result = 0;
    u64 addr = (u64)(*ppos) - MVPCI_IOW_SEQUENCER_OFFSET;
    struct mvpci_fh *fh = (struct mvpci_fh *)file->private_data;
    struct mvpci_driver *mvpci_drv = fh->mvpci_drv;

    if ((addr + count) <= MVPCI_IOW_SEQUENCER_SIZE) {
        u8 *p = (u8 *)mvpci_drv->iow_sequencer + addr;
        result = copy_to_user(buf, p, count);
        if (result) {
            pr_err("ERROR: read from addr=0x%llx, count=0x%zx, result=0x%x.\n", addr, count, result);
            return -EINVAL;
        }
    } else { 
        pr_err("ERROR: read length (0x%zx) invalid. Must be <= 0x%llx.\n", count, MVPCI_IOW_SEQUENCER_SIZE);
        return -EINVAL;
    }
#if 0
    dprintk(2, "%s read from addr=0x%llx, count=0x%zx, result=0x%x\n", __FUNCTION__, addr, count, result);
#endif
    return result ? result : count;
}

static ssize_t mvpci_read(struct file *file, char __user *buf, size_t count,
              loff_t *ppos)
{
#define MAX_READ_LEN (4 * _kB)
    int result = 0;
    u64 addr = (u64)(*ppos);
    u64 val = 0;
    struct mvpci_fh *fh = (struct mvpci_fh *)file->private_data;
    struct mvpci_driver *mvpci_drv = fh->mvpci_drv;

    if (addr >= MVPCI_IOW_SEQUENCER_OFFSET) {
        /* reading from io_writer sequencer memory */
        return mvpci_read_seq(file, buf, count, ppos);
    }

    if (addr > pci_resource_len(mvpci_drv->pci, 0)) {
        pr_err("ERROR: Offset 0x%llx is outside range of BAR 0 (0x%llx).\n",  addr, pci_resource_len(mvpci_drv->pci, 0));
        return -ENOMEM;
    }

    if (count == sizeof(u32)) {
        val = mvread(addr);
        result = copy_to_user(buf, &val, sizeof(u32));
    } else if (count == sizeof(u64)) {
        val = mvread64(addr);
        result = copy_to_user(buf, &val, sizeof(u64));
    } else if (count <= MAX_READ_LEN) {
        unsigned char *val_buf = vmalloc(MAX_READ_LEN);
        if (val_buf) {
            mvreadbulk((void *)val_buf, addr, count);
            result = copy_to_user(buf, val_buf, count);
            vfree(val_buf);
        } else {
            pr_err("ERROR: cannot allocate memory for bulk read.");
            return -ENOMEM;
        }
    } else { 
        pr_err("ERROR: read length (0x%zx) invalid. Must be <= 0x%x.", count, MAX_READ_LEN);
        return -EINVAL;
    }
#if 0
    dprintk(2, "%s read from addr=0x%llx, val=0x%llx, count=0x%zx, result=0x%x\n", __FUNCTION__, addr, val, count, result);
#endif
    return result ? result : count;
}

static int read_maxpayload(struct pci_dev *dev)
{
    int pos;
    u16 dctl, dcap, dsz, dmax;

    pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
    if (!pos)
        return -1;

    /* Read Device MaxPayload capability and current setting */
    pci_read_config_word(dev, pos + PCI_EXP_DEVCTL, &dctl);
    pci_read_config_word(dev, pos + PCI_EXP_DEVCAP, &dcap);
    dsz = (dctl & PCI_EXP_DEVCTL_PAYLOAD) >> 0x05;
    dmax = (dcap & PCI_EXP_DEVCAP_PAYLOAD);
    // only 3 values are possible (0:128, 1:256, 2:512 bytes)
    if ((dsz < 0x03) && (dmax < 0x03))
    {
        dprintk(0, "MaxPayload: used = %d Bytes, capability = %d Bytes\n",
                0x80 * (0x1 << dsz), 0x80 * (0x1 << dmax));
        return 0x80 * (0x1 << dsz);
    }
    dprintk(0, "MaxPayload: *UNKNOWN*!\n");
    return -1;
}

static int read_maxreadreq(struct pci_dev *dev)
{
    int pos;
    u16 dctl, dsz;

    pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
    if (!pos)
        return -1;

    /* Read Device MaxReadReq current setting */
    pci_read_config_word(dev, pos + PCI_EXP_DEVCTL, &dctl);
    dsz = (dctl & PCI_EXP_DEVCTL_READRQ) >>  0x0C;
    // only 6 values are possible (0:128, 1:256, 2:512, 3:1024, 4:2048, 5:4096 bytes) but we only allow <= 1024 to be used
    if (dsz > 0x03)
    {
        dsz = 0x3;
    }
    dprintk(0, "MaxReadReq: using %d Bytes\n", 0x80 * (0x1 << dsz));
    return 0x80 * (0x1 << dsz);
}

#ifndef PCI_EXP_LNKSTA_CLS_16_0GB
#  define PCI_EXP_LNKSTA_CLS_16_0GB 0x0004 /* Current Link Speed 16.0GT/s */
#endif

#ifndef PCI_EXP_LNKSTA_CLS_32_0GB
#  define PCI_EXP_LNKSTA_CLS_32_0GB 0x0005 /* Current Link Speed 32.0GT/s */
#endif

static u32 get_pcie_link_speed(struct mvpci_driver *mvpci_drv)
{
	u16 linkstat;
	u32 speed = 2500;               /* assume Gen 1 */
	int ret;

	ret = pcie_capability_read_word(mvpci_drv->pci, PCI_EXP_LNKSTA, &linkstat);
	if (ret) {
        dprintk(0, "*** Unable to read the PCIe configuration! Assume PCIe Gen1.\n");
		return speed;
	}

    switch (linkstat & PCI_EXP_LNKSTA_CLS)
    {
    case PCI_EXP_LNKSTA_CLS_5_0GB:
        speed = 5000;               /* Gen 2, 5GHz */
        break;
    case PCI_EXP_LNKSTA_CLS_8_0GB:
        speed = 8000;               /* Gen 3, 8GHz */
        break;
    case PCI_EXP_LNKSTA_CLS_16_0GB:
        speed = 16000;              /* Gen 4, 16GHz */
        break;
    case PCI_EXP_LNKSTA_CLS_32_0GB:
        speed = 32000;              /* Gen ?, 32GHz */
        break;
    case PCI_EXP_LNKSTA_CLS_2_5GB:
    default:
        speed = 2500;               /* Gen 1, 2.5GHz */
        break;
    }

	return speed;
}

static u32 get_pcie_link_width(struct mvpci_driver *mvpci_drv)
{
	u16 linkstat;
	int ret;

	ret = pcie_capability_read_word(mvpci_drv->pci, PCI_EXP_LNKSTA, &linkstat);
	if (ret) {
        dprintk(0, "*** Unable to read the PCIe configuration! Assume x1.\n");
		return 1;   /* assume Gen1 */
	}

	return (linkstat & PCI_EXP_LNKSTA_NLW) >> PCI_EXP_LNKSTA_NLW_SHIFT;
}

#define CONVERT_PCIE_SPEED_TO_MBS(speed) \
      ((speed) == 32000 ? 32000*128/130 : \
      (speed) == 16000 ? 16000*128/130 : \
      (speed) == 8000  ?  8000*128/130 : \
      (speed) == 5000  ?  5000*8/10 : \
      (speed) == 2500  ?  2500*8/10 : \
      0)

static u32 get_pcie_link_bandwidth(struct mvpci_driver *mvpci_drv)
{
    u32 link_width;
    u32 link_speed;
    u32 link_bandwidth;

    link_speed = get_pcie_link_speed(mvpci_drv);
    link_width = get_pcie_link_width(mvpci_drv);
    link_bandwidth = link_width * CONVERT_PCIE_SPEED_TO_MBS(link_speed);
    dprintk(1, "link bandwidth: %d Mb/s (width = %d, bus speed = %d)\n",
        link_bandwidth, link_width, link_speed);

    return link_bandwidth;
 }

static long mvpci_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    long result = 0;
    u32 cp_size;
    u32 io_size = _IOC_SIZE(cmd);
    u32 arg32;
    struct mvpci_fh *fh = (struct mvpci_fh *)filp->private_data;
    struct mvpci_driver *mvpci_drv = fh->mvpci_drv;

    if (fh) {
        switch (cmd) {
            case MVPCI_IOC_WRITE_BAR: {
                TmvPCI_Register32 reg;
                cp_size = CHECK_IO_SIZE( io_size, sizeof(TmvPCI_Register32));
                if ((result = copy_from_user( &reg, (TmvPCI_Register32 __user *)arg, cp_size))) {
                    result = -EFAULT;
                    break;
                }

                dprintk(
                    MVPCI_FLAG_IORW,
                    "device mvpci-%d: MVPCI_IOC_WRITE_BAR: offset=0x%04x value=0x%04x size=%ld bytes\n",
                    mvpci_drv->nr, reg.offset, reg.value, reg.size);
                mvwrite( reg.value, reg.offset );
                break;
            }
            case MVPCI_IOC_READ_BAR: {
                TmvPCI_Register32 reg;
                long retval = 0;
                cp_size = CHECK_IO_SIZE( io_size, sizeof(TmvPCI_Register32));
                       if ((result = copy_from_user( &reg, (TmvPCI_Register32 __user *)arg, cp_size))) {
                    result = -EFAULT;
                    break;
                }

                retval = mvread( reg.offset );

                if (retval < 0) {
                    result = retval;
                    break;
                }
                reg.value = retval;
                dprintk(
                    MVPCI_FLAG_IORW,
                    "device mvpci-%d: MVPCI_IOC_READ_BAR: offset=0x%04x value=0x%04x size=%ld bytes\n",
                    mvpci_drv->nr, reg.offset, reg.value, reg.size);

                retval = copy_to_user((TmvPCI_Register32 __user *)arg, &reg, sizeof(reg));
                break;
            }
            case MVPCI_IOC_READ_MAXPAYLOAD: {
                dprintk(1, "device mvpci-%d: MVPCI_READ_MAXPAYLOAD: %d\n", mvpci_drv->nr, mvpci_drv->maxpayloadsize);
                result = copy_to_user((u32 __user *) arg, &mvpci_drv->maxpayloadsize, sizeof(u32));
                break;
            }
            case MVPCI_IOC_READ_MAXREADREQ: {
                dprintk(1, "device mvpci-%d: MVPCI_READ_MAXREADREQ: %d\n", mvpci_drv->nr, mvpci_drv->maxreadreq);
                result = copy_to_user((u32 __user *) arg, &mvpci_drv->maxreadreq, sizeof(u32));
                break;
            }
            case MVPCI_IOC_READ_FPGA_FAMILY: {
                dprintk(1, "device mvpci-%d: MVPCI_READ_FPGA_FAMILY: %d\n", mvpci_drv->nr, mvpci_drv->revision);
                /* only the lower 8 bits are significant, the others should be set to 0 */
                arg32 = (u32)mvpci_drv->revision;
                result = copy_to_user((u32 __user *) arg, &arg32, sizeof(u32));
                break;
            }
            case MVPCI_IOC_READ_FPGA_VERSION: {
                dprintk(1, "device mvpci-%d: MVPCI_READ_FPGA_VERSION: %d\n", mvpci_drv->nr, mvpci_drv->subsystem_id);
                /* only the lower 16 bits are significant, the others should be set to 0 */
                arg32 = (u32)mvpci_drv->subsystem_id;
                result = copy_to_user((u32 __user *) arg, &arg32, sizeof(u32));
                break;
            }
            case MVPCI_IOC_READ_IOWSEQ_BASE_ADDR: {
                dprintk(1, "device mvpci-%d: MVPCI_IOC_READ_IOWSEQ_BASE_ADDR: %pad\n", mvpci_drv->nr, &mvpci_drv->dma_iow_sequencer);
                result = copy_to_user((u64 __user *) arg, &mvpci_drv->dma_iow_sequencer, sizeof(u64));
                break;
            }
            case MVPCI_IOC_READ_PCIE_LINK_WIDTH: {
                arg32 = get_pcie_link_width(mvpci_drv);
                dprintk(1, "device mvpci-%d: MVPCI_IOC_READ_PCIE_WIDTH: %d\n", mvpci_drv->nr, arg32);
                result = copy_to_user((u32 __user *) arg, &arg32, sizeof(u32));
                break;
            }
            case MVPCI_IOC_READ_PCIE_BUS_SPEED: {
                arg32 = get_pcie_link_speed(mvpci_drv);
                dprintk(1, "device mvpci-%d: MVPCI_IOC_READ_PCIE_BUS_SPEED: %d\n", mvpci_drv->nr, arg32);
                result = copy_to_user((u32 __user *) arg, &arg32, sizeof(u32));
                break;
            }
            case MVPCI_IOC_READ_PCIE_BANDWIDTH: {
                arg32 = get_pcie_link_bandwidth(mvpci_drv);
                dprintk(1, "device mvpci-%d: MVPCI_IOC_READ_PCIE_BANDWIDTH: %d\n", mvpci_drv->nr, arg32);
                result = copy_to_user((u32 __user *) arg, &arg32, sizeof(u32));
               break;
            }
            default: {
                result = mvdma_ioctl(filp, cmd, arg);    /* use the dma "device's" ioctl */
                break;
            }
        }
    }
    return result;
}

static const struct file_operations mvpci_fops = {
    .owner = THIS_MODULE,
    .open = mvpci_open,
    .release = mvpci_release,
    .write = mvpci_write,
    .read = mvpci_read,
    .unlocked_ioctl = mvpci_ioctl,
};

static int camera_reset_write(int camera_num)
{
    int retval = -1;
    if (camera_num < number_of_devices) {
        struct mvpci_driver* mvpci_drv = &mvpci_devices[camera_num];

        if (mvpci_drv && (mvpci_drv->nr == camera_num) && (mvpci_drv->initialized != 0)) {
            struct pci_dev* our_device = mvpci_drv->pci;
            if (mvpci_drv->fh->opened) {
#ifdef COWARDLY_REFUSAL
                dprintk(0, "device mvpci-%d: cowardly refusing to reset a device that is being used!\n", mvpci_drv->nr);
                return retval;
#else
                dprintk(0, "device mvpci-%d: device is open %d times, but we are going to ignore it!\n", mvpci_drv->nr, mvpci_drv->fh->opened);
#endif // COWARDLY_REFUSAL
            }
            if (our_device) {
                struct pci_bus* our_bus = our_device->bus;
                struct pci_dev* our_parent = our_bus->self;
                struct pci_bus* our_parents_bus = our_parent->bus;

                dprintk(0, "device mvpci-%d: our camera device (0x%04x:0x%04x) at %s is subordinate to the bridge (0x%04x:0x%04x) at %s\n",
                    mvpci_drv->nr,
                    our_device->vendor, our_device->device, pci_name(our_device),
                    our_parent->vendor, our_parent->device, pci_name(our_parent));

                // remove the device and signal to the km that it has gone
                // this will make mvpci_drv invalid, so do not use beyond this point!
                pci_stop_and_remove_bus_device(our_device);
                dprintk(0, "resetting the secondary bus...\n");
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 19, 0)
                pci_reset_bridge_secondary_bus(our_parent);
#else
                pci_bridge_secondary_bus_reset(our_parent);
#endif
                dprintk(0, "rescanning the secondary bus...\n");
                pci_rescan_bus(our_parents_bus);
                retval = 0;
            }
        }
    }
	return retval;
}

#define DEFINE_CAMERA_RESET(n)                               \
static ssize_t camera_reset_write ## n(struct kobject *kobj, struct kobj_attribute *attr, \
			 const char *buf, size_t count)                 \
{                                                           \
    camera_reset_write(n);                                  \
    return count;                                           \
}

DEFINE_CAMERA_RESET(0)
DEFINE_CAMERA_RESET(1)
DEFINE_CAMERA_RESET(2)
DEFINE_CAMERA_RESET(3)
#define RESET_CAMERA(n) camera_reset_write ## n

#define DEFINE_RESET_ATTR(n)                               \
static struct kobj_attribute camera_reset_attribute ## n = \
	__ATTR(camera_reset ## n, 0220, NULL, RESET_CAMERA(n));

DEFINE_RESET_ATTR(0)
DEFINE_RESET_ATTR(1)
DEFINE_RESET_ATTR(2)
DEFINE_RESET_ATTR(3)

#define DEFINE_CAMERA_RESET_ATTR(n)                         \
static struct attribute *attrs ## n[] = {                   \
	&camera_reset_attribute ## n.attr,                      \
	NULL,                                                   \
};

DEFINE_CAMERA_RESET_ATTR(0)
DEFINE_CAMERA_RESET_ATTR(1)
DEFINE_CAMERA_RESET_ATTR(2)
DEFINE_CAMERA_RESET_ATTR(3)

/*
 * An unnamed attribute group will put all of the attributes directly in
 * the kobject directory.  If we specify a name, a subdirectory will be
 * created for the attributes with the directory being the name of the
 * attribute group.
 */

#define DEFINE_CAMERA_RESET_GROUP(n)                        \
static struct attribute_group attr_group ## n = {           \
	.attrs = attrs ## n,                                    \
};

DEFINE_CAMERA_RESET_GROUP(0)
DEFINE_CAMERA_RESET_GROUP(1)
DEFINE_CAMERA_RESET_GROUP(2)
DEFINE_CAMERA_RESET_GROUP(3)

static void mvsysfs_init(struct mvpci_driver *mvpci_drv)
{
	int err;

    /* Create sysfs entry and associate it with the device */
    switch (mvpci_drv->nr) {
        case 0:
            err = sysfs_create_group(mvpci_kobj, &attr_group0);
            break;
        case 1:
            err = sysfs_create_group(mvpci_kobj, &attr_group1);
            break;
        case 2:
            err = sysfs_create_group(mvpci_kobj, &attr_group2);
            break;
        case 3:
            err = sysfs_create_group(mvpci_kobj, &attr_group3);
            break;
        default:
            err = 1;
            break;
    }
}

static void mvsysfs_deinit(struct mvpci_driver *mvpci_drv)
{
    // remove group
    switch (mvpci_drv->nr) {
        case 0:
            sysfs_remove_group(&(mvpci_drv->pci->dev.kobj), &attr_group0);
            break;
        case 1:
            sysfs_remove_group(&(mvpci_drv->pci->dev.kobj), &attr_group1);
            break;
        case 2:
            sysfs_remove_group(&(mvpci_drv->pci->dev.kobj), &attr_group2);
            break;
        case 3:
            sysfs_remove_group(&(mvpci_drv->pci->dev.kobj), &attr_group3);
            break;
        default:
            break;
    }
}

static int mvpci_probe(struct pci_dev *pci_dev,
               const struct pci_device_id *pci_id)
{
    int i;
    int result;
    unsigned char lat;
    struct mvpci_driver *mvpci_drv;
    int num_vectors = 0;
    int mvpci_num = 0;

    if (number_of_devices >= MVPCI_MAX_DEVICES)
        return -ENOMEM;

    for (i = 0; i < MVPCI_MAX_DEVICES; i++) {
        mvpci_drv = &mvpci_devices[i];
        /* find the first free device structure - we might be reinitialising after a PCI rescan */
        if (mvpci_drv->initialized == 0) {
            mvpci_num = i;
            break;
        }
    }
    if (i >= MVPCI_MAX_DEVICES)
        return -ENOMEM;

    dprintk(1,
        "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
    dprintk(0, "MVPCI PCI device %d found.\n", mvpci_num);

    /* clear everything! */
    memset(mvpci_drv, 0, sizeof(*mvpci_drv));

    mvpci_drv->nr = mvpci_num;
    mvpci_drv->pci = pci_dev;
    mvpci_drv->id = pci_dev->device;
    mvpci_drv->flags = 0;
    mvpci_drv->initialized = 0;
    mvpci_drv->fh[0].mvpci_drv = mvpci_drv;
    sprintf(mvpci_drv->name, "mvpci-%d", mvpci_drv->nr);

#if defined(CONFIG_NUMA) && defined(USE_NUMA)
    /* try to get the node number for this device */
    mvpci_drv->node = pcibus_to_node(mvpci_drv->pci->bus);
    if (mvpci_drv->node < first_node(node_online_map))
        mvpci_drv->node = first_node(node_online_map);
#endif /* CONFIG_NUMA */

    mvpci_drv->devnum =
        MKDEV(MAJOR(mvpci_devnum), MINOR(mvpci_devnum) + mvpci_drv->nr);
    dprintk(0, "MAJOR=%d, minor=%d\n", MAJOR(mvpci_devnum),
        MINOR(mvpci_devnum) + mvpci_drv->nr);

    if (pci_enable_device(mvpci_drv->pci)) {
        dprintk(0, "can't enable device %d\n", mvpci_drv->nr);
        mvpci_drv->nr = -1;
        return -EIO;
    }
    pci_set_master(mvpci_drv->pci);
    pci_set_drvdata(mvpci_drv->pci, mvpci_drv);

    if (!pci_set_dma_mask(mvpci_drv->pci, DMA_BIT_MASK(64))) {
        dprintk(0, "64 bit DMA available on PCIe device %s.\n",
            mvpci_drv->name);
    } else {
        dprintk(0, "Error! 64 bit DMA NOT available on PCIe device %s.\n",
            mvpci_drv->name);
        mvpci_drv->nr = -1;
        return -EIO;
    }

#ifdef USE_COHERENT_DMA
    if (!dma_set_coherent_mask(&mvpci_drv->pci->dev, DMA_BIT_MASK(64))) {
        dprintk(0, "Coherent DMA is available on PCIe device %s.\n",
            mvpci_drv->name);
        mvpci_drv->coherent = 1;
    } else {
        dprintk(0, "Warning! Coherent DMA is NOT available on PCIe device %s.\n",
            mvpci_drv->name);
    }
#endif

    if (pci_resource_flags(mvpci_drv->pci, 0) & IORESOURCE_MEM) {
        dprintk(0, "found MMIO @ BAR-0 of length 0x%llx on device %d\n",
                pci_resource_len(mvpci_drv->pci, 0),
                mvpci_drv->nr);
    } else {
        dprintk(0, "no MMIO available on device %d\n", mvpci_drv->nr);
        mvpci_drv->nr = -1;
        return -EIO;
    }

    if (pci_request_regions(mvpci_drv->pci, mvpci_drv->name)) {
        dprintk(0, "can't request iomem for device %d\n",
            mvpci_drv->nr);
        mvpci_drv->nr = -1;
        return -EBUSY;
    } else {
        mvpci_drv->flags |= MVPCI_FLAG_RES0;
    }

    mvpci_drv->maxpayloadsize = read_maxpayload(mvpci_drv->pci);
    mvpci_drv->maxreadreq = read_maxreadreq(mvpci_drv->pci);

    pci_read_config_byte(mvpci_drv->pci, PCI_REVISION_ID,
                 &mvpci_drv->revision);
    pci_read_config_word(mvpci_drv->pci, PCI_SUBSYSTEM_ID,
                 &mvpci_drv->subsystem_id);
    dprintk(0, "device 0x%04x (revision: 0x%x, subsystem_id: 0x%04x) at %s\n",
        mvpci_drv->id, mvpci_drv->revision, mvpci_drv->subsystem_id,
        pci_name(mvpci_drv->pci));

    dprintk(0, "link bandwidth: %d Mb/s (width = %d, bus speed = %d)\n",
        get_pcie_link_bandwidth(mvpci_drv), get_pcie_link_width(mvpci_drv), get_pcie_link_speed(mvpci_drv));

    pci_read_config_byte(mvpci_drv->pci, PCI_LATENCY_TIMER, &lat);
    dprintk(1, "non msi irq: %d, latency: %d, resource_start: 0x%llx\n",
        mvpci_drv->pci->irq, lat,
        (unsigned long long)pci_resource_start(mvpci_drv->pci, 0));

    mvpci_drv->mmio = pci_iomap(mvpci_drv->pci, 0, pci_resource_len(mvpci_drv->pci, 0));
    if (mvpci_drv->mmio == NULL) {
        dprintk(0, "ERROR: device %d pci_iomap failed\n",
            mvpci_drv->nr);
        result = -EIO;
        goto fail;
    }
    dprintk(1, "mapped %d kB register space to 0x%px\n",
        (u32)(pci_resource_len(mvpci_drv->pci, 0) >> 10),
        mvpci_drv->mmio);

    mvpci_drv->iow_sequencer = dma_alloc_coherent(&mvpci_drv->pci->dev,
            MVPCI_IOW_SEQUENCER_SIZE, &mvpci_drv->dma_iow_sequencer, GFP_DMA); /* this is the only flag allowed here */
    if (!mvpci_drv->iow_sequencer) {
        dprintk(0, "ERROR: device %d reserving io_writer sequencer memory failed\n",
            mvpci_drv->nr);
        result = -EIO;
        goto fail;
    }
    dprintk(1, "reserved %lld kB for the io_writer sequencer @ 0x%pK/%pad\n",
        MVPCI_IOW_SEQUENCER_SIZE / _kB,
        mvpci_drv->iow_sequencer,
        &mvpci_drv->dma_iow_sequencer);

    result = mvpci_init_hw(mvpci_drv);

    if (result < 0) {
        dprintk(0, "ERROR: init device %d hw failed (%d)\n", mvpci_num,
            result);
        goto fail;
    }

    for (i = 0; i < MVPCI_MSI_MAX_VECTORS; i++) {
        mvpci_drv->msi_x_entry[i].entry = i;
    }

    /* disable Interrupts from the hardware */
    mvwrite(0x00, MVPCI_IO_CONTROL);

#if defined(TRY_MSIX) || defined(TRY_MSI)
    /* try to use MSI-X or MSI interrupts */
    dprintk(0, "trying to use %d vector(s) on device mvpci-%d\n", MVPCI_MSI_MAX_VECTORS, mvpci_num);
#ifdef TRY_MSIX
    num_vectors = pci_msix_vec_count(mvpci_drv->pci);
    if (num_vectors > 0) {
        dprintk(0, "support for %d MSI-X vectors on device mvpci-%d\n", num_vectors, mvpci_num);
    }
#endif /* TRY_MSIX */
#ifdef TRY_MSI
    num_vectors = pci_msi_vec_count(mvpci_drv->pci);
    if (num_vectors > 0) {
        dprintk(0, "support for %d MSI vectors on device mvpci-%d\n", num_vectors, mvpci_num);
    }
#endif /* TRY_MSI */
#endif //defined(TRY_MSIX) || defined(TRY_MSI)

#if defined(TRY_MSIX)
    result = pci_alloc_irq_vectors(mvpci_drv->pci, MVPCI_MSI_MIN_VECTORS, MVPCI_MSI_MAX_VECTORS, PCI_IRQ_MSI | PCI_IRQ_MSIX | PCI_IRQ_AFFINITY);
#elif defined(TRY_MSI)
    result = pci_alloc_irq_vectors(mvpci_drv->pci, MVPCI_MSI_MIN_VECTORS, MVPCI_MSI_MAX_VECTORS, PCI_IRQ_MSI | PCI_IRQ_AFFINITY);
#else
    result = -1;
#endif /* TRY_MSI */

    if (result < 0) {
        /* only LEGACY INTERRUPTS will work */
        result = pci_alloc_irq_vectors(mvpci_drv->pci, 1, 1, PCI_IRQ_LEGACY);
        if (result < 0) {
            dprintk(0, "ERROR: device mvpci-%d: failed to obtain a single legacy interrupt for MSeq!\n", mvpci_num);
            goto fail;
        }
        num_vectors = result;
        dprintk(0, "device mvpci-%d: legacy interrupts only will be used on this system.\n", mvpci_num);
        mvpci_drv->flags |= MVPCI_FLAG_IRQ;
    } else {
        /* MSI and MSI-X return 1 or more vectors */
        num_vectors = result;
        dprintk(0, "device mvpci-%d: %d MSI or MSI-X interrupt vectors supplied on this system.\n", mvpci_num, num_vectors);
        if ((num_vectors != MVPCI_MSI_MIN_VECTORS) && (num_vectors != MVPCI_MSI_MAX_VECTORS)) {
            dprintk(0, "ERROR: device mvpci-%d: system supplies %d interrupts, but the driver needs either 1 or %d!\n", mvpci_num, result, MVPCI_MSI_MAX_VECTORS);
            goto fail;
        }
        mvpci_drv->flags |= MVPCI_FLAG_MSI;
    }

    for (i = 0; i < MVPCI_MSI_MAX_VECTORS; i++) {
        irq_handler_t handler = NULL;
        int uio_interrupt = 0;
        char uio_dev_name[10] = "";

        result = -1;
        switch (i) {
    case INT_MSEQ:
            if (num_vectors == 1) 
                /* legacy interrupt  or just 1 MSI available */
                handler = mvpci_isr_single;
            else
                /* >1 MSI or MS-X interrupt */
                handler = mvpci_isr_mseq;
            break;
    case INT_IO:
            uio_interrupt = 1;
            strcpy(uio_dev_name, "mvio");
            handler = (irq_handler_t)MSI_UIO_ISR(INT_IO);
            break;
    case INT_STAT:
            uio_interrupt = 1;
            strcpy(uio_dev_name, "mvstat");
            handler = (irq_handler_t)MSI_UIO_ISR(INT_STAT);
            break;
    case INT_SC:
            uio_interrupt = 1;
            strcpy(uio_dev_name, "mvsc");
            handler = (irq_handler_t)MSI_UIO_ISR(INT_SC);
            break;
    default:
            break;
        }

        if (uio_interrupt) {
            const char *name;

            /* register a UIO interrupt handler */
            result = set_allocate_name(&name, "%s:mvpci%d", uio_dev_name, mvpci_num);
            if (result) {
                dprintk(0, "ERROR: can't set uio interrupt name for device %d, vector %d, IRQ %d\n",
                    mvpci_num, i,
                    mvpci_drv->msi_x_entry[i].vector);
                goto fail;
            }

            mvpci_drv->uio_info[i].name = name;
            mvpci_drv->uio_info[i].version = MVPCI_VERSION_STR;
            if (num_vectors == 1) {
                dprintk(0, "device mvpci-%d: set uio interrupt name %s, vector %d, *CUSTOM* IRQ\n",
                        mvpci_num, name, i);
                mvpci_drv->msi_x_entry[i].vector = UIO_IRQ_CUSTOM;
                mvpci_drv->uio_info[i].irq = UIO_IRQ_CUSTOM;
                mvpci_drv->uio_info[i].priv = (void *)mvpci_drv;
            } else {
                dprintk(0, "device mvpci-%d: set uio interrupt name %s, vector %d, IRQ %d\n",
                        mvpci_num, name, i,
                        mvpci_drv->msi_x_entry[i].vector);
                mvpci_drv->msi_x_entry[i].vector = pci_irq_vector(mvpci_drv->pci, i);
                mvpci_drv->uio_info[i].irq = mvpci_drv->msi_x_entry[i].vector;
                mvpci_drv->uio_info[i].irq_flags = IRQF_SHARED;
                mvpci_drv->uio_info[i].handler = (irq_uio_handler_t)handler;
                mvpci_drv->uio_info[i].priv = (void *)mvpci_drv;
            }

            result = uio_register_device(
                        &pci_dev->dev,
                        &mvpci_drv->uio_info[i]);

            if (result < 0) {
                dprintk(0, "ERROR: device mvpci-%d: uio_register_device failed using vector %d, IRQ %d, returns %d\n",
                    mvpci_num, i,
                    mvpci_drv->msi_x_entry[i].vector, result);
                /* tidy up */
                mvpci_drv->msi_x_entry[i].vector = 0;
                mvpci_drv->uio_info[i].name = "";
                kfree_const(name);
                goto fail;
            }
            /* else OK */

        } else if (handler) {
            /* register a regular interrupt handler */
            mvpci_drv->msi_x_entry[i].vector = pci_irq_vector(mvpci_drv->pci, i);
            if ((i == INT_MSEQ) && mvpci_drv->flags & MVPCI_FLAG_IRQ)
            {
                dprintk(0, "device mvpci-%d trying to use a single legacy interrupt %d for MSeq.\n", mvpci_num, mvpci_drv->msi_x_entry[INT_MSEQ].vector);
                mvpci_drv->pci->irq = mvpci_drv->msi_x_entry[INT_MSEQ].vector;
            } else {
                dprintk(0, "device mvpci-%d trying to use a MSI interrupt %d for MSeq.\n", mvpci_num, mvpci_drv->msi_x_entry[i].vector);
            }

            result = request_irq(
                mvpci_drv->msi_x_entry[i].vector,
                handler, IRQF_SHARED, mvpci_drv->name,
                (void *)mvpci_drv);
        } else {
            /* unused interrupt - ignore */
            continue;
        }

        if (result < 0) {
            dprintk(0, "ERROR: device mvpci-%d: vector %d, IRQ %d\n",
                mvpci_num, i,
                mvpci_drv->msi_x_entry[i].vector);
            goto fail;
        } else {
            dprintk(0, "device mvpci-%d: using vector %d, IRQ %d\n",
                mvpci_num, i,
                mvpci_drv->msi_x_entry[i].vector);
        }
    } /* for */

    /* setup character device */
    cdev_init(&mvpci_drv->cdev, &mvpci_fops);
    if ((result = cdev_add(&mvpci_drv->cdev, mvpci_drv->devnum, 1))) {
        dprintk(0, "ERROR: device %d cdev_add failed %d\n", mvpci_drv->nr,
            result);
        goto fail;
    } else {
        dprintk(1, "device %d cdev_add succeeded\n", mvpci_drv->nr);
    }

    /* setup dma 'device' */
    result = mvdma_probe(mvpci_drv);
    if (result < 0) {
        dprintk(0, "ERROR: can't setup DMA structures\n");
        goto fail;
    } else {
        dprintk(1, "Setup DMA structures OK\n");
    }

    number_of_devices++;
    mvpci_drv->initialized = 1;
    dprintk(0, "initialised mvpci device %d (total devices = %d)\n", mvpci_drv->nr, number_of_devices);
    return 0;

fail:
    if (mvpci_drv->iow_sequencer) {
        dma_free_coherent(&mvpci_drv->pci->dev, MVPCI_IOW_SEQUENCER_SIZE, mvpci_drv->iow_sequencer, mvpci_drv->dma_iow_sequencer);
    }
    mvpci_free_pci(mvpci_drv);
    return result;
}

static void mvpci_remove(struct pci_dev *pci_dev)
{
    struct mvpci_driver *mvpci_drv = pci_get_drvdata(pci_dev);
    int nr = mvpci_drv->nr;

    dprintk(1, "removing device %d, name = %s, dev = %s...\n", nr, mvpci_drv->name, pci_name(pci_dev));
    mvpci_free_pci(mvpci_drv);
    if (number_of_devices > 0)
        number_of_devices--;
    dprintk(0, "device %d removed, %d devices remain\n", nr, number_of_devices);

    return;
}


#ifdef USE_PCI_ERROR_HANDLERS
static pci_ers_result_t mvpci_error_detected(struct pci_dev *pci_dev,
                    pci_channel_state_t state)
{
#if 0 // TODO
    struct mvpci_driver *mvpci_drv = pci_get_drvdata(pci_dev);
#endif
    switch (state) {
    case pci_channel_io_normal:
        return PCI_ERS_RESULT_CAN_RECOVER;
    case pci_channel_io_frozen:
        pr_warn("frozen state error, reset controller, going offline\n");
#if 0 // TODO
        mvpci_device_offline(pci_dev, mvpci_drv);
#endif
        pci_disable_device(pci_dev);
        return PCI_ERS_RESULT_NEED_RESET;
    case pci_channel_io_perm_failure:
        pr_warn("failure state error, req. disconnect\n");
        return PCI_ERS_RESULT_DISCONNECT;
    }
    return PCI_ERS_RESULT_NEED_RESET;
}

static pci_ers_result_t mvpci_slot_reset(struct pci_dev *pci_dev)
{
#if 0 // TODO
    struct mvpci_driver *mvpci_drv = pci_get_drvdata(pci_dev);
#endif
    pr_info("restart after slot reset\n");
    if (pci_enable_device_mem(pci_dev)) {
        pr_info("failed to renable after slot reset\n");
        return PCI_ERS_RESULT_DISCONNECT;
    }

    pci_set_master(pci_dev);
    pci_restore_state(pci_dev);
    pci_save_state(pci_dev);
#if 0 // TODO
    mvpci_device_online(pci_dev, mvpci_drv);
#endif

    return PCI_ERS_RESULT_RECOVERED;
}

static void mvpci_error_resume(struct pci_dev *pci_dev)
{
    pr_info("error resume\n");
#if 0 // TODO
    pci_cleanup_aer_uncorrect_error_status(pci_dev);
#endif
}

static void mvpci_reset_prepare(struct pci_dev *pci_dev)
{
#if 0 // TODO
    struct mvpci_driver *mvpci_drv = pci_get_drvdata(pci_dev);
#endif
    pr_info("reset prepare, going offline\n");
#if 0 // TODO
    mvpci_device_offline(pci_dev, mvpci_drv);
#endif
}

static void mvpci_reset_done(struct pci_dev *pci_dev)
{
#if 0 // TODO
    struct mvpci_driver *mvpci_drv = pci_get_drvdata(pci_dev);
#endif
    pr_info("reset done, going online\n");
#if 0 // TODO
    mvpci_device_online(pci_dev, mvpci_drv);
#endif
}

static const struct pci_error_handlers mvpci_err_handler = {
    .error_detected    = mvpci_error_detected,
    .slot_reset    = mvpci_slot_reset,
    .resume        = mvpci_error_resume,
    .reset_prepare    = mvpci_reset_prepare,
    .reset_done    = mvpci_reset_done,
};
#endif /* USE_PCI_ERROR_HANDLERS */


#ifndef PCI_VENDOR_ID_MV
#define PCI_VENDOR_ID_MV 0x4d56 /* MV */
#endif
#define PCI_ID_PRODUCT1 0x7014
#define PCI_ID_PRODUCT2 0x7024
#define PCI_ID_PRODUCT3 0x7012
#define PCI_ID_PRODUCT4 0x7022
static struct pci_device_id mvpci_pci_tbl[] = {
    {PCI_VENDOR_ID_MV, 0x0000, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
    {PCI_VENDOR_ID_MV, PCI_ID_PRODUCT1, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
    {PCI_VENDOR_ID_MV, PCI_ID_PRODUCT2, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
    {PCI_VENDOR_ID_MV, PCI_ID_PRODUCT3, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
    {PCI_VENDOR_ID_MV, PCI_ID_PRODUCT4, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
    {
        0,
    }};

MODULE_DEVICE_TABLE(pci, mvpci_pci_tbl);

static struct pci_driver mvpci_pci_driver = {
    .name = "mvpci",
    .id_table = mvpci_pci_tbl,
    .probe = mvpci_probe,
    .remove = mvpci_remove,
#ifdef USE_PCI_ERROR_HANDLERS
    .err_handler = &mvpci_err_handler,
#endif /* USE_PCI_ERROR_HANDLERS */
};

static int mvpci_init_module(void)
{
    unsigned nr;

#ifndef CONFIG_PCI_MSI
    dprintk(0, "ERROR: Kernel needs to be configured with CONFIG_PCI_MSI!\n");
    return -EFAULT;
#endif /* CONFIG_PCI_MSI */

    memset(mvpci_devices, 0,
        MVPCI_MAX_DEVICES * sizeof(struct mvpci_driver));

    if (alloc_chrdev_region(&mvpci_devnum, 0, MVPCI_MAX_DEVICES, "mvpci") == 0) {
        dprintk(0, "Successfully registered mvpci as a character device\n");
    } else {
        pr_err("%s: Can't register mvpci as a character device\n", __func__);
        return -EFAULT;
    }

    dprintk(0, "mvpci v%s loaded (major %d)\n", MVPCI_VERSION_STR,
        MAJOR(mvpci_devnum));
    dprintk(0, "PAGE_SIZE on this system is 0x%08lx\n", PAGE_SIZE);


    /* our sysfs entries go here */
    mvpci_kobj = kobject_create_and_add("matrix-vision", kernel_kobj);

    if (pci_register_driver(&mvpci_pci_driver) == 0) {
        dprintk(0, "Successfully registered mvpci as a PCI driver\n");
    } else {
        pr_err("%s: Can't register mvpci as a PCI driver\n", __func__);
        return -EFAULT;
    }

    // create sysfs entries for devices that we have found
    for (nr = 0; nr < number_of_devices; nr++) {
        struct mvpci_driver *mvpci_drv = &mvpci_devices[nr];
        mvsysfs_init(mvpci_drv);
    }
    return 0;
}

static void mvpci_cleanup_module(void)
{
    unsigned nr;

    for (nr = 0; nr < number_of_devices; nr++) {
        struct mvpci_driver *mvpci_drv = &mvpci_devices[nr];

        if (mvpci_drv->nr >= 0) {
            // remove sysfs entries for devices that exist at the moment
            mvsysfs_deinit(mvpci_drv);
            // remove character devices
            cdev_del(&mvpci_drv->cdev);
        }
    }

    // remove our sysfs directory
    if (mvpci_kobj)
        kobject_put(mvpci_kobj);

    unregister_chrdev_region(mvpci_devnum, MVPCI_MAX_DEVICES);
    pci_unregister_driver(&mvpci_pci_driver);
    dprintk(0, "Unregistered mvPCI driver\n");
    return;
}

module_init(mvpci_init_module);
module_exit(mvpci_cleanup_module);
